from typing import List

from moodle.attr import dataclass, field
from .notification_processor import NotificationProcessor


@dataclass
class NotificationComponent:
    displayname: str
    preferencekey: str
    processors: List[NotificationProcessor] = field(factory=list)



from moodle.base.general import GeneralTypeValue


class CategoryType(GeneralTypeValue):
    """Cohort category / type

    Args:
        type (str): the name of the field: id (numeric value
                        of course category id) or idnumber (alphanumeric value of idnumber course category)
                        or system (value ignored)
        value (str): the value of the categorytype
    """
    pass



from typing import Optional

from moodle.attr import dataclass
from . import CategoryType


@dataclass
class UpdateCohort:
    """arg for core_cohort_update_cohorts

    Args:
        id (str): ID of the cohort
        categorytype (CategoryType): Cohort category / type
        name (str): cohort name
        idnumber (str): cohort idnumber
        description (Optional[str]): cohort description
        descriptionformat (int): Default to "1" description format (1 = HTML, 0 = MOODLE, 2 = PLAIN or 4 = MARKDOWN)
        visible (Optional[int]): cohort visible
        theme (Optional[str]): the cohort theme. The allowcohortthemes setting must be enabled on Moodle
    """
    id: str
    categorytype: CategoryType
    name: str
    idnumber: str
    description: Optional[str] = None
    descriptionformat: int = 1
    visible: Optional[int] = None
    theme: Optional[str] = None



from datetime import datetime
from typing import List, Optional, Union

from moodle import BaseMoodle
from . import (
    AccessInformation,
    AllowedEventTypes,
    CalendarExportToken,
    Event,
    Events,
    CourseEvents,
    ActionEventCourses,
    DayView,
    MonthlyView,
    Timestamps,
)


class BaseCalendar(BaseMoodle):
    def create_calendar_events(self, events: List[Events.Create]) -> Events:
        """Create calendar events

        Args:
            events (List[Events.Create]): list of Events.Create

        Returns:
            Events: list of Event
        """
        res = self.moodle.post(
            'core_calendar_create_calendar_events',
            events=events,
        )
        return self._tr(Events, **res)

    def delete_calendar_events(self, events: List[Events.Delete]) -> None:
        """Delete calendar events

        Args:
            events (List[Events.Delete]): list of Events.Delete or list of {eventid:int, repeat:int}
        """
        self.moodle.post('core_calendar_delete_calendar_events', events=events)

    def get_action_events_by_course(self,
                                    courseid: int,
                                    timesortfrom: Union[datetime, int] = None,
                                    timesortto: Union[datetime, int] = None,
                                    aftereventid: int = 0,
                                    limitnum: int = 20) -> CourseEvents:
        """Get calendar action events by course

        Args:
            courseid (int): Course id
            timesortfrom (Union[datetime, int], optional): Time sort from. Defaults to None.
            timesortto (Union[datetime, int], optional): Time sort to. Defaults to None.
            aftereventid (int, optional): The last seen event id. Defaults to 0.
            limitnum (int, optional): Limit number. Defaults to 20.

        Returns:
            CourseEvents: list of CourseEvent
        """
        res = self.moodle.post(
            'core_calendar_get_action_events_by_course',
            courseid=courseid,
            timesortfrom=timesortfrom,
            timesortto=timesortto,
            aftereventid=aftereventid,
            limitnum=limitnum,
        )
        return self._tr(CourseEvents, **res)

    def get_action_events_by_courses(self,
                                     courseids: List[int],
                                     timesortfrom: Union[datetime, int] = None,
                                     timesortto: Union[datetime, int] = None,
                                     limitnum: int = 10) -> ActionEventCourses:
        """Get calendar action events by courses

        Args:
            courseids (List[int]): list of Course id
            timesortfrom (Union[datetime, int], optional): Time sort from. Defaults to None.
            timesortto (Union[datetime, int], optional): Time sort to. Defaults to None.
            limitnum (int, optional): Limit number. Defaults to 10.

        Returns:
            ActionEventCourses: list of CourseEvents
        """
        res = self.moodle.post(
            'core_calendar_get_action_events_by_courses',
            courseids=courseids,
            timesortfrom=timesortfrom,
            limitnum=limitnum,
        )
        return self._tr(ActionEventCourses, **res)

    def get_action_events_by_timesort(self,
                                      timesortfrom: Union[datetime, int] = 0,
                                      timesortto: Union[datetime, int] = None,
                                      aftereventid: int = 0,
                                      limitnum: int = 20,
                                      limittononsuspendedevents: int = None,
                                      userid: int = None) -> CourseEvents:
        """Get calendar action events by tiemsort

        Args:
            timesortfrom (Union[datetime, int], optional): Time sort from. Defaults to 0.
            timesortto (Union[datetime, int], optional): Time sort to. Defaults to None.
            aftereventid (int, optional): The last seen event id. Defaults to 0.
            limitnum (int, optional): Limit number. Defaults to 20.
            limittononsuspendedevents (int, optional): Limit the events to courses the user is not suspended in. Defaults to None.
            userid (int, optional): The user id. Defaults to None.

        Returns:
            CourseEvents: list of CourseEvent
        """
        res = self.moodle.post(
            'core_calendar_get_action_events_by_timesort',
            timesortfrom=timesortfrom,
            timesortto=timesortto,
            aftereventid=aftereventid,
            limitnum=limitnum,
            limittononsuspendedevents=limittononsuspendedevents,
            userid=userid,
        )
        return self._tr(CourseEvents, **res)

    def get_allowed_event_types(self, courseid: int = 0) -> AllowedEventTypes:
        """Get the type of events a user can create in the given course.

        Args:
            courseid (int, optional): Course to check, empty for site.. Defaults to 0.

        Returns:
            AllowedEventTypes: list of allowed event types to be created in the given course.
        """
        res = self.moodle.post(
            'core_calendar_get_allowed_event_types',
            courseid=courseid,
        )
        return self._tr(AllowedEventTypes, **res)

    def get_calendar_access_information(self,
                                        courseid: int = 0
                                        ) -> AccessInformation:
        """Convenience function to retrieve some permissions/access information for the given course calendar.

        Args:
            courseid (int, optional): Course to check, empty for site calendar events. Defaults to 0.

        Returns:
            AccessInformation: Permissions/access information for the given course calendar
        """
        res = self.moodle.post(
            'core_calendar_get_calendar_access_information',
            courseid=courseid,
        )
        return self._tr(AccessInformation, **res)

    def get_calendar_day_view(self,
                              year: int,
                              month: int,
                              day: int,
                              courseid: int = 1,
                              categoryid: Optional[int] = None) -> DayView:
        """Fetch the day view data for a calendar

        Args:
            year (int): Year to be viewed
            month (int): Month to be viewed
            day (int): Day to be viewed
            courseid (int, optional): Course being viewed. Defaults to 1.
            categoryid (Optional[int], optional): Category being viewed. Defaults to None.

        Returns:
            DayView: Response list of Event
        """
        res = self.moodle.post(
            'core_calendar_get_calendar_day_view',
            year=year,
            month=month,
            day=day,
            courseid=courseid,
            categoryid=categoryid,
        )
        return self._tr(DayView, **res)

    def get_calendar_event_by_id(self, eventid: int) -> CourseEvents:
        """Get calendar event by id

        Args:
            eventid (int): The event id to be retrieved

        Returns:
            CourseEvents: list of Event
        """
        res = self.moodle.post(
            'core_calendar_get_calendar_event_by_id',
            eventid=eventid,
        )
        return self._tr(CourseEvents, **res)

    def get_calendar_events(self,
                            events: Events.Details = None,
                            options: Events.Options = None) -> Events:
        """Get calendar events

        Args:
            events (Events.Details, optional): Event details. Defaults to None.
            options (Events.Options, optional): Options. Defaults to None.

        Returns:
            Events: list of Event
        """
        res = self.moodle.post(
            'core_calendar_get_calendar_events',
            events=events if events else {},
            options=options if options else {},
        )
        return self._tr(Events, **res)

    def get_calendar_export_token(self) -> CalendarExportToken:
        """Return the auth token required for exporting a calendar.

        Returns:
            CalendarExportToken: The calendar permanent access token for calendar export.
        """
        res = self.moodle.post('core_calendar_get_calendar_export_token')
        return self._tr(CalendarExportToken, **res)

    def get_calendar_monthly_view(self,
                                  year: int,
                                  month: int,
                                  courseid: int = 1,
                                  categoryid: int = None,
                                  includenavigation: int = 1,
                                  mini: int = None,
                                  day: int = 1) -> MonthlyView:
        """Fetch the monthly view data for a calendar

        Args:
            year (int): Year to be viewed
            month (int): Month to be viewed
            courseid (int, optional): Course being viewed. Defaults to 1.
            categoryid (int, optional): Category being viewed. Defaults to None.
            includenavigation (int, optional): Whether to return the mini month view or not. Defaults to 1.
            mini (int, optional): Whether to return the mini month view or not. Defaults to None.
            day (int, optional): Day to be viewed. Defaults to 1.

        Returns:
            MonthlyView: [description]
        """
        res = self.moodle.post(
            'core_calendar_get_calendar_monthly_view',
            year=year,
            month=month,
            courseid=courseid,
            categoryid=categoryid,
            includenavigation=includenavigation,
            mini=mini,
            day=day,
        )
        return self._tr(MonthlyView, **res)

    def get_calendar_upcoming_view(self,
                                   courseid: int = 1,
                                   categoryid: int = None) -> CourseEvents:
        """Fetch the upcoming view data for a calendar

        Args:
            courseid (int): Course being viewed. Defaults to 1.
            categoryid (int, optional): Category being viewed. Defaults to None.

        Returns:
            CourseEvents: [description]
        """
        res = self.moodle.post(
            'core_calendar_get_calendar_upcoming_view',
            courseid=courseid,
            categoryid=categoryid,
        )
        return self._tr(CourseEvents, **res)

    def get_timestamps(self, data: List[Timestamps.Create]) -> Timestamps:
        """Fetch unix timestamps for given date times.

        Args:
            data (List[Timestamps.Create]): list of Timestamps.Create

        Returns:
            Timestamps: list of Timestamp
        """
        res = self.moodle.post('core_calendar_get_timestamps', data=data)
        return self._tr(Timestamps, **res)

    def submit_create_update_form(self,
                                  formdata: str) -> CourseEvents.EventForm:
        """Submit form data for event form

        Args:
            formdata (str): The data from the event form

        Returns:
            CourseEvents.EventForm: Event form
        """
        res = self.moodle.post(
            'core_calendar_submit_create_update_form',
            formdata=formdata,
        )
        return self._tr(CourseEvents.EventForm, **res)

    def update_event_start_day(self, eventid: int,
                               daytimestamp: Union[datetime, int]) -> Event:
        """Update the start day (but not time) for an event.

        Args:
            eventid (int): Id of event to be updated
            daytimestamp (Union[datetime, int]): Timestamp for the new start day

        Returns:
            Event: Updated event
        """
        res = self.moodle.post(
            'core_calendar_update_event_start_day',
            eventid=eventid,
            daytimestamp=daytimestamp,
        )
        return Event(**res['event'])  # type: ignore



from typing import List

from moodle import MoodleWarning, ResponsesFactory
from moodle.attr import dataclass, field


@dataclass
class AllowedEventTypes(ResponsesFactory[str]):
    """The type of events a user can create in the given course.

    Args:
        allowedeventtypes (List[str]): list of allowed event type
        warnings (List[Warning]): list of MoodleWarning
    """
    allowedeventtypes: List[str] = field(factory=list)
    warnings: List[MoodleWarning] = field(factory=list)



from typing import List

from moodle import MoodleWarning
from moodle.attr import dataclass, field


@dataclass
class GeneralValidated:
    """GeneralValidated
    Args:
        validated (int): True or False, nor 1 or 0
        warnings (List[Warning]): list of warnings
    """
    validated: int
    warnings: List[MoodleWarning] = field(factory=list)

    def __bool__(self) -> bool:
        if isinstance(self.validated, int):
            return self.validated == 1
        elif isinstance(self.validated, str):
            return self.validated == '1' or self.validated == 'true'
        return bool(self.validated)



from typing import List, Optional

from moodle import ResponsesFactory
from moodle.attr import dataclass, field


@dataclass
class Cohort:
    """Cohort

    Args:
        id (int): ID of the cohort
        name (str): cohort name
        idnumber (str): cohort idnumber
        description (str): cohort description
        descriptionformat (int): description format (1 = HTML, 0 = MOODLE, 2 = PLAIN or 4 = MARKDOWN)
        visible (int): cohort visible
        theme (Optional[str]): cohort theme
    """
    id: int
    name: str
    idnumber: str
    description: str
    descriptionformat: int
    visible: int
    theme: Optional[str]


@dataclass
class Cohorts(ResponsesFactory[Cohort]):
    """List of Cohort
    """
    cohorts: List[Cohort] = field(factory=list)

    @property
    def items(self) -> List[Cohort]:
        return self.cohorts



from moodle.attr import dataclass


@dataclass
class DeleteCohort:
    """arg for core_cohort_delete_cohort_members

    Args:
        cohortid (int): cohort record id
        userid (int): user id
    """
    cohortid: int
    userid: int



from typing import Optional, List

from moodle import ResponsesFactory, MoodleWarning
from moodle.attr import dataclass, field


@dataclass
class Comment:
    """Comment

    Args:
        id (int): Comment ID
        content (str): The content text formatted
        format (int): content format (1 = HTML, 0 = MOODLE, 2 = PLAIN or 4 = MARKDOWN)
        timecreated (int): Time created (timestamp)
        strftimeformat (str): Time format
        profileurl (str): URL profile
        fullname (str): fullname
        time (str): Time in human format
        avatar (str): HTML user picture
        userid (int): User ID
        delete (Optional[int]): Permission to delete=true/false
    """
    id: int
    content: str
    format: int
    timecreated: int
    strftimeformat: str
    profileurl: str
    fullname: str
    time: str
    avatar: str
    userid: int
    delete: Optional[int]


@dataclass
class Comments(ResponsesFactory[Comment]):
    """List of Comments

    Args:
        comments (List[Comment]): list of Comments
        count (Optional[int]): Total number of comments.
        perpage (Optional[int]): Number of comments per page.
        canpost (Optional[int]): Whether the user can post in this comment area.
        warnings (List[MoodleWarning]): list of MoodleWarnings
    """
    comments: List[Comment] = field(factory=list)
    count: Optional[int] = None
    perpage: Optional[int] = None
    canpost: Optional[int] = None
    warnings: List[MoodleWarning] = field(factory=list)

    @property
    def items(self) -> List[Comment]:
        return self.comments

    @dataclass
    class Create:
        """Adds a comment or comments.

        Args:
            contextlevel (str): contextlevel system, course, user...
            instanceid (int): the id of item associated with the contextlevel
            component (str): component
            content (str): component
            itemid (int): associated id
            area (str, optional): string comment area. Default to ""
        """
        contextlevel: str
        instanceid: int
        component: str
        content: str
        itemid: int
        area: str = ""



from moodle import MoodleWarning
from typing import List
from moodle.attr import dataclass, field


@dataclass
class CalendarExportToken:
    """Auth token for exporting a calendar.
        token (str): The calendar permanent access token for calendar export.
        warnings (List[MoodleWarning]): list of MoodleWarning
    """
    token: str
    warnings: List[MoodleWarning] = field(factory=list)

    def __str__(self):
        return self.token



from moodle.attr import dataclass
from . import CohortType, UserType


@dataclass
class AddCohortMember:
    """Adds cohort members arg

    Args:
        cohorttype (CohortType): Cohort type
        usertype (UserType): Cohort Usertype
    """
    cohorttype: CohortType
    usertype: UserType



from moodle.base.general import GeneralTypeValue


class CohortType(GeneralTypeValue):
    """Cohort type

    Args:
        type (str): The name of the field: id
                        (numeric value of cohortid) or idnumber (alphanumeric value of idnumber)
        value (str): The value of the cohort
    """
    pass



import cattr
from typing import Any, Iterable, List, Type, TypeVar, TYPE_CHECKING
if TYPE_CHECKING:
    from moodle import Moodle

T = TypeVar('T')


class BaseMoodle(object):
    def __init__(self, moodle):
        self._moodle = moodle
        self.__post_init__(moodle)

    def __post_init__(self, moodle) -> None:
        pass

    @property
    def moodle(self) -> 'Moodle':
        return self._moodle

    @staticmethod
    def _tr(kls: Type[T], *args, **kwargs) -> T:
        return cattr.structure(kwargs, kls)  # type: ignore

    @staticmethod
    def _trs(kls: Type[T], datas: Iterable[Any]) -> List[T]:
        return [cattr.structure(data, kls)
                for data in datas] if datas else []  # type: ignore



from typing import List

from moodle import MoodleWarning
from moodle.attr import dataclass, field


@dataclass
class ViewNotes:
    """Return type for view notes
    Constructor arguments:
    params: status (int): status: true if success
    params: warnings (List[Warning]): status: list of warnings
    """
    status: int
    warnings: List[MoodleWarning] = field(factory=list)



from typing import List

from moodle import MoodleWarning
from moodle.attr import dataclass, field


@dataclass
class GeneralStatus:
    """GeneralStatus
    Args:
        success (int): True or False, nor 1 or 0
        warnings (List[Warning]): list of warnings
    """
    status: int
    warnings: List[MoodleWarning] = field(factory=list)



from attr import attrs, attrib, has, Attribute
from attr import asdict as asdict_attr
from datetime import datetime
from functools import partial
from typing import Any, Callable, List

COMMON: List = [str, int, datetime]


def serialize(inst=None,
              field: Attribute = None,
              value: Any = None,
              name: str = None) -> Any:
    if not value:
        return value
    if not name:
        name = getattr(field, 'name', '')
    if isinstance(value, list):
        out = {}
        for idx, val in enumerate(value):
            val = serialize(value=val)
            if isinstance(val, dict):
                for key, value in val.items():
                    out[f'{name}[{idx}][{key}]'] = val[key]
            else:
                out_key = name or ''
                # Check if data required name prefix
                if hasattr(value, 'name'):
                    out_key += f"[{getattr(value, 'name')}]"
                out_key += f'[{idx}]'
                out[out_key] = val
        return out
    elif isinstance(value, dict):
        out = {}
        for key, value in value.items():
            if isinstance(value, list):
                out.update(serialize(value=val, name=key))
            else:
                out[key] = value
        return out
    elif has(value):
        return asdict_attr(value, value_serializer=serialize)  # type: ignore
    elif isinstance(value, datetime):
        return datetime.timestamp(value)
    return value


asdict: Callable = partial(
    asdict_attr,
    value_serializer=serialize,
)
dataclass: Callable = partial(
    attrs,
    auto_attribs=True,
    slots=True,
)
field = attrib



from typing import List

from moodle import BaseMoodle, MoodleWarning
from . import Comment, Comments


class BaseComment(BaseMoodle):
    def add_comments(self, comments: List[Comments.Create]) -> List[Comment]:
        """Adds a comment or comments.

        Args:
            comments (List[Comments.Create]): list of Comments.Create

        Returns:
            List[Comment]: list of Comment
        """
        data = self.moodle.get('core_comment_add_comments', comments=comments)
        return self._trs(Comment, data)

    def delete_comments(self, comments: List[int]) -> List[MoodleWarning]:
        """Deletes a comment or comments.

        Args:
            comments (List[int]): list id of the comment

        Returns:
            List[MoodleWarning]: list of warnings
        """
        data = self.moodle.get(
            'core_comment_delete_comments',
            comments=comments,
        )
        return self._trs(MoodleWarning, data)

    def get_comments(self,
                     contextlevel: str,
                     instanceid: int,
                     component: int,
                     itemid: int,
                     area: str = "",
                     page: int = 0,
                     sortdirection: str = "DESC") -> Comments:
        """Returns comments.

        Args:
            contextlevel (str): contextlevel system, course, user...
            instanceid (int): the Instance id of item associated with the context level
            component (int): component
            itemid (int): associated id
            area (str, optional): string comment area. Defaults to "".
            page (int, optional): page number (0 based). Defaults to 0.
            sortdirection (str, optional): Sort direction: ASC or DESC. Defaults to "DESC".

        Returns:
            Comments: Response
        """
        data = self.moodle.get(
            'core_comment_get_comments',
            contextlevel=contextlevel,
            instanceid=instanceid,
            component=component,
            itemid=itemid,
            area=area,
            page=page,
            sortdirection=sortdirection,
        )
        return self._tr(Comments, data)



from .category_type import CategoryType
from .create_cohort import CreateCohort
from .update_cohort import UpdateCohort
from .delete_cohort import DeleteCohort
from .cohort_members import CohortMembers
from .cohort_type import CohortType
from .user_type import UserType
from .add_member import AddCohortMember
from .cohort import Cohort, Cohorts

from .base import BaseCohort

__all__ = [
    'CategoryType',
    'CreateCohort',
    'UpdateCohort',
    'DeleteCohort',
    'CohortMembers',
    'CohortType',
    'UserType',
    'AddCohortMember',
    'Cohort',
    'Cohorts',
    'BaseCohort',
]



from .access_information import AccessInformation
from .allowed_event_types import AllowedEventTypes
from .event import Event, Events
from .calendar_export_token import CalendarExportToken
from .course_event import (CourseEventIcon, CourseEventCategory,
                           CourseEventCourse, CourseEventSubscription,
                           CourseEventAction, CourseEvent, CourseEvents,
                           ActionEventCourses)
from .period import Period
from .date_view import DateView
from .day_view import DayView
from .monthly_view import Day, Week, DayName, MonthlyView
from .timestamp import Timestamp, Timestamps

from .base import BaseCalendar

__all__ = [
    'AccessInformation',
    'AllowedEventTypes',
    'Event',
    'Events',
    'CalendarExportToken',
    'CourseEventIcon',
    'CourseEventCategory',
    'CourseEventCourse',
    'CourseEventSubscription',
    'CourseEventAction',
    'CourseEvent',
    'CourseEvents',
    'ActionEventCourses',
    'Period',
    'DateView',
    'DayView',
    'Day',
    'Week',
    'DayName',
    'MonthlyView',
    'Timestamp',
    'Timestamps',
    'BaseCalendar',
]



from typing import List, Optional

from moodle import BaseMoodle
from . import Grades


class BaseGrades(BaseMoodle):
    def create_gradecategory(
        self,
        courseid: int,
        fullname: str,
        aggregation: Optional[int] = None,
        aggregateonlygraded: Optional[int] = None,
        aggregateoutcomes: Optional[int] = None,
        droplow: Optional[int] = None,
        itemname: Optional[str] = None,
        iteminfo: Optional[str] = None,
        idnumber: Optional[str] = None,
        gradetype: Optional[int] = None,
        grademax: Optional[int] = None,
        grademin: Optional[int] = None,
        gradepass: Optional[int] = None,
        display: Optional[int] = None,
        decimals: Optional[int] = None,
        hiddenuntil: Optional[int] = None,
        locktime: Optional[int] = None,
        weightoverride: Optional[int] = None,
        aggregationcoef2: Optional[str] = None,
        parentcategoryid: Optional[int] = None,
        parentcategoryidnumber: Optional[str] = None,
    ) -> Optional[int]:
        """Create a grade category inside a course gradebook.

        Args:
            courseid (int): id of course
            fullname (str): fullname of category
            aggregation (Optional[int], optional): aggregation method. Defaults to None.
            aggregateonlygraded (Optional[int], optional): exclude empty grades. Defaults to None.
            aggregateoutcomes (Optional[int], optional): aggregate outcomes. Defaults to None.
            droplow (Optional[int], optional): drop low grades. Defaults to None.
            itemname (Optional[str], optional): the category total name. Defaults to None.
            iteminfo (Optional[str], optional): the category iteminfo. Defaults to None.
            idnumber (Optional[str], optional): the category idnumber. Defaults to None.
            gradetype (Optional[int], optional): the grade type. Defaults to None.
            grademax (Optional[int], optional): the grade max. Defaults to None.
            grademin (Optional[int], optional): the grade min. Defaults to None.
            gradepass (Optional[int], optional): the grade to pass. Defaults to None.
            display (Optional[int], optional): the display type. Defaults to None.
            decimals (Optional[int], optional): the decimal count. Defaults to None.
            hiddenuntil (Optional[int], optional): grades hidden until. Defaults to None.
            locktime (Optional[int], optional): lock grades after. Defaults to None.
            weightoverride (Optional[int], optional): weight adjusted. Defaults to None.
            aggregationcoef2 (Optional[str], optional): weight coefficient. Defaults to None.
            parentcategoryid (Optional[int], optional): The parent category id. Defaults to None.
            parentcategoryidnumber (Optional[str], optional): the parent category idnumber. Defaults to None.

        Returns:
            Optional[int]: The ID of the created category
        """
        options = dict(
            aggregation=aggregation,
            aggregateonlygraded=aggregateonlygraded,
            aggregateoutcomes=aggregateoutcomes,
            droplow=droplow,
            itemname=itemname,
            iteminfo=iteminfo,
            idnumber=idnumber,
            gradetype=gradetype,
            grademax=grademax,
            grademin=grademin,
            gradepass=gradepass,
            display=display,
            decimals=decimals,
            hiddenuntil=hiddenuntil,
            locktime=locktime,
            weightoverride=weightoverride,
            aggregationcoef2=aggregationcoef2,
            parentcategoryid=parentcategoryid,
            parentcategoryidnumber=parentcategoryidnumber,
        )
        # TODO : Refactor options of core_grades_create_gradecategory
        data = self.moodle.post(
            'core_grades_create_gradecategory',
            courseid=courseid,
            fullname=fullname,
            options=options,
        )
        return dict(data).get('categoryid')

    def get_grades(self,
                   courseid: int,
                   component: int = None,
                   activityid: int = None,
                   userids: List[int] = None) -> Grades:
        """**DEPRECATED** Please do not call this function any more. Returns student course total grade and grades for activities. This function does not return category or manual items. This function is suitable for managers or teachers not students.

        Args:
            courseid (int): id of course
            component (int, optional): A component, for example mod_forum or mod_quiz. Defaults to None.
            activityid (int, optional): The activity ID. Defaults to None.
            userids (List[int], optional): An array of user IDs, leave empty to just retrieve grade item information. Defaults to None.

        Returns:
            Grades: Student course total grade and grades for activities.
        """
        data = self.moodle.post(
            'core_grades_get_grades',
            component=component or '',
            activityid=activityid,
            userids=userids or list(),
        )
        return self._tr(Grades, data)

    def grader_gradingpanel_point_fetch(
        self,
        component: str,
        contextid: int,
        itemname: str,
        gradeduserid: int,
    ) -> dict:
        """Fetch the data required to display the grader grading panel for simple grading, creating the grade item if required

        Args:
            component (str): The name of the component
            contextid (int): The ID of the context being graded
            itemname (str): The grade item itemname being graded
            gradeduserid (int): The ID of the user show

        Returns:
            dict: gradingpanel
        """
        data = self.moodle.post(
            'core_grades_grader_gradingpanel_point_fetch',
            component=component,
            contextid=contextid,
            itemname=itemname,
            gradeduserid=gradeduserid,
        )
        # TODO : Add type for core_grades_grader_gradingpanel_point_fetch!
        return data

    def grader_gradingpanel_point_store(
        self,
        component: str,
        contextid: int,
        itemname: str,
        gradeduserid: int,
        formdata: str,
        notifyuser: int = None,
    ) -> dict:
        """Store the data required to display the grader grading panel for simple grading

        Args:
            component (str): The name of the component
            contextid (int): The ID of the context being graded
            itemname (str): The grade item itemname being graded
            gradeduserid (int): The ID of the user show
            formdata (str): The serialised form data representing the grade
            notifyuser (int, optional): Wheteher to notify the user or not. Defaults to None.

        Returns:
            dict: gradingpanel
        """
        data = self.moodle.post(
            'core_grades_grader_gradingpanel_point_store',
            component=component,
            contextid=contextid,
            itemname=itemname,
            gradeduserid=gradeduserid,
            notifyuser=notifyuser or '',
            formdata=formdata,
        )
        # TODO : Add type for core_grades_grader_gradingpanel_point_store!
        return data

    def grader_gradingpanel_scale_fetch(
        self,
        component: str,
        contextid: int,
        itemname: str,
        gradeduserid: int,
    ) -> dict:
        """Fetch the data required to display the grader grading panel for scale-based grading, creating the grade item if required

        Args:
            component (str): The name of the component
            contextid (int): The ID of the context being graded
            itemname (str): The grade item itemname being graded
            gradeduserid (int): The ID of the user show

        Returns:
            dict: gradingpanel scale
        """
        data = self.moodle.post(
            'core_grades_grader_gradingpanel_scale_fetch',
            component=component,
            contextid=contextid,
            itemname=itemname,
            gradeduserid=gradeduserid,
        )
        # TODO : Add type for core_grades_grader_gradingpanel_scale_fetch!
        return data

    def grader_gradingpanel_scale_store(
        self,
        component: str,
        contextid: int,
        itemname: str,
        gradeduserid: int,
        formdata: str,
        notifyuser: int = None,
    ):
        """Store the data required to display the grader grading panel for scale-based grading

        Args:
            component (str): The name of the component
            contextid (int): The ID of the context being graded
            itemname (str): The grade item itemname being graded
            gradeduserid (int): The ID of the user show
            formdata (str): The serialised form data representing the grade
            notifyuser (int, optional): Wheteher to notify the user or not. Defaults to None.

        Returns:
            [type]: [description]
        """
        data = self.moodle.post(
            'core_grades_grader_gradingpanel_scale_store',
            component=component,
            contextid=contextid,
            itemname=itemname,
            gradeduserid=gradeduserid,
            notifyuser=notifyuser or '',
            formdata=formdata,
        )
        # TODO : Add type for core_grades_grader_gradingpanel_scale_store!
        return data

    def update_grades(
        self,
        source: str,
        courseid: int,
        component: str,
        activityid: int,
        itemnumber: int,
        grades: List[dict],
        itemname: Optional[str] = None,
        idnumber: Optional[int] = None,
        gradetype: Optional[int] = None,
        grademax: Optional[float] = None,
        grademin: Optional[float] = None,
        scaleid: Optional[int] = None,
        multfactor: Optional[float] = None,
        plusfactor: Optional[float] = None,
        deleted: Optional[int] = None,
        hidden: Optional[int] = None,
    ) -> int:
        """Update a grade item and associated student grades.

        Args:
            source (str): The source of the grade update
            courseid (int): id of course
            component (str): A component, for example mod_forum or mod_quiz
            activityid (int): The activity ID
            itemnumber (int): grade item ID number for modules that have multiple grades. Typically this is 0.
            grades (List[dict]): Any student grades to alter. list of {studentid:int, grade:float, str_feedback: Optional[str]}
            itemname (Optional[str], optional): The grade item name. Defaults to None.
            idnumber (Optional[int], optional): Arbitrary ID provided by the module responsible for the grade item. Defaults to None.
            gradetype (Optional[int], optional): The type of grade (0 = none, 1 = value, 2 = scale, 3 = text). Defaults to None.
            grademax (Optional[float], optional): Maximum grade allowed. Defaults to None.
            grademin (Optional[float], optional): Minimum grade allowed. Defaults to None.
            scaleid (Optional[int], optional): The ID of the custom scale being is used. Defaults to None.
            multfactor (Optional[float], optional): Multiply all grades by this number. Defaults to None.
            plusfactor (Optional[float], optional): Add this to all grades. Defaults to None.
            deleted (Optional[int], optional): True if the grade item should be deleted. Defaults to None.
            hidden (Optional[int], optional): True if the grade item is hidden. Defaults to None.

        Returns:
            int: A value like 0 => OK, 1 => FAILED as defined in lib/grade/constants.php
        """
        itemdetails = dict(
            itemname=itemname,
            idnumber=idnumber,
            gradetype=gradetype,
            grademax=grademax,
            grademin=grademin,
            scaleid=scaleid,
            multfactor=multfactor,
            plusfactor=plusfactor,
            deleted=deleted,
            hidden=hidden,
        )
        # TODO : Refactor itemdetails of core_grades_create_gradecategory
        data = self.moodle.post(
            'core_grades_update_grades',
            source=source,
            courseid=courseid,
            component=component,
            activityid=activityid,
            itemnumber=itemnumber,
            grades=grades,
            itemdetails=itemdetails,
        )
        return data



from typing import List

from moodle.attr import dataclass, field


@dataclass
class Grade:
    """Grade

    Args:
        userid (int): Student ID
        grade (float): Student grade
        locked (int): 0 means not locked, > 1 is a date to lock until
        hidden (int): 0 means not hidden, 1 hidden, > 1 is a date to hide until
        overridden (int): 0 means not overridden, > 1 means overridden
        feedback (str): Feedback from the grader
        feedbackformat (int): The format of the feedback
        usermodified (int): The ID of the last user to modify this student grade
        datesubmitted (int): A timestamp indicating when the student submitted the activity
        dategraded (int): A timestamp indicating when the assignment was grades
        str_grade (str): A string representation of the grade
        str_long_grade (str): A nicely formatted string representation of the grade
        str_feedback (str): A formatted string representation of the feedback from the grader
    """
    userid: int
    grade: float
    locked: int
    hidden: int
    overridden: int
    feedback: str
    feedbackformat: int
    usermodified: int
    datesubmitted: int
    dategraded: int
    str_grade: str
    str_long_grade: str
    str_feedback: str


@dataclass
class GradeItem:
    """Item of Grades

    Args:
        activityid (str): The ID of the activity or "course" for the course grade item
        itemnumber (int): Will be 0 unless the module has multiple grades
        scaleid (int): The ID of the custom scale or 0
        name (str): The module name
        grademin (float): Minimum grade
        grademax (float): Maximum grade
        gradepass (float): The passing grade threshold
        locked (int): 0 means not locked, > 1 is a date to lock until
        hidden (int): 0 means not hidden, > 1 is a date to hide until
        grades (List[Grade]): list of Grade
    """
    activityid: str
    itemnumber: int
    scaleid: int
    name: str
    grademin: float
    grademax: float
    gradepass: float
    locked: int
    hidden: int
    grades: List[Grade] = field(factory=list)


@dataclass
class OutCome:
    """OutCome of Grades

    Args:
        activityid (str): The ID of the activity or "course" for the course grade item
        itemnumber (int): Will be 0 unless the module has multiple grades
        scaleid (int): The ID of the custom scale or 0
        name (str): The module name
        locked (int): 0 means not locked, > 1 is a date to lock until
        hidden (int): 0 means not hidden, > 1 is a date to hide until
        grades (List[Grade]): list of Grade
    """
    activityid: str
    itemnumber: int
    scaleid: int
    name: str
    locked: int
    hidden: int
    grades: List[Grade] = field(factory=list)


@dataclass
class Grades:
    """Grades

    Args:
        items (List[GradeItem]): An array of items associated with the grade items
        outcomes (List[OutCome]): An array of outcomes associated with the grade items
    """
    items: List[GradeItem] = field(factory=list)
    outcomes: List[OutCome] = field(factory=list)



from typing import List

from moodle import MoodleWarning
from moodle.attr import dataclass, field


@dataclass
class View:
    """View Folder
    Args:
        status (int): status: true if success
        warnings (List[Warning]): list of warnings
    """
    status: int
    warnings: List[MoodleWarning] = field(factory=list)

    def __bool__(self) -> bool:
        if type(self.status) == bool:
            return bool(self.status)
        return self.status == 1



from typing import List

from moodle import ResponsesFactory
from moodle.attr import dataclass, field


@dataclass
class CohortMembers(ResponsesFactory[int]):
    """Cohort's Members

    Args:
        cohortid (int): cohort record id
        userids (List[int]): list of user id
    """
    cohortid: int
    userids: List[int] = field(factory=list)

    @property
    def items(self) -> List[int]:
        return self.userids



from typing import List

from moodle import MoodleWarning
from moodle.attr import dataclass, field


@dataclass
class SignupUserResponse:
    """SignupUser
    Args:
        success (int): True if the user was created false otherwise
        warnings (List[Warning]): list of warnings
    """
    success: int
    warnings: List[MoodleWarning] = field(factory=list)



from typing import List

from moodle import BaseMoodle, MoodleWarning
from . import (
    Cohort,
    Cohorts,
    CreateCohort,
    UpdateCohort,
    DeleteCohort,
    CohortMembers,
    AddCohortMember,
)


class BaseCohort(BaseMoodle):
    def add_cohort_members(
            self, members: List[AddCohortMember]) -> List[MoodleWarning]:
        """Adds cohort members.

        Args:
            members (List[AddCohortMember]): list of AddCohortMember

        Returns:
            List[MoodleWarning]: list of MoodleWarning, if any
        """
        data = self.moodle.get(
            'core_cohort_add_cohort_members',
            members=members,
        )
        warnings = dict(data).get('warnings', list())
        warnings = warnings if warnings else list()
        return self._trs(MoodleWarning, data)

    def create_cohorts(self, cohorts: List[CreateCohort]) -> List[Cohort]:
        """Creates new cohorts.

        Args:
            cohorts (List[CreateCohort]): list of CreateCohort

        Returns:
            List[Cohort]: list of Cohort
        """
        data = self.moodle.get('core_cohort_create_cohorts', cohorts=cohorts)
        return self._trs(Cohort, data)

    def delete_cohort_members(self, members: List[DeleteCohort]) -> None:
        """Deletes cohort members.

        Args:
            members (List[DeleteCohort]): list of DeleteCohort or {cohortid:int, userid:int}
        """
        self.moodle.get('core_cohort_delete_cohort_members', members=members)

    def delete_cohorts(self, cohortids: List[int]) -> None:
        """Deletes all specified cohorts.

        Args:
            cohortids (List[int]): list of cohort ID
        """
        self.moodle.get('core_cohort_delete_cohorts', cohortids=cohortids)

    def get_cohort_members(self, cohortids: List[int]) -> List[CohortMembers]:
        """Returns cohort members.

        Args:
            cohortids (List[int]): list of Cohort ID

        Returns:
            List[CohortMembers]: list of CohortMembers
        """
        data = self.moodle.get(
            'core_cohort_get_cohort_members',
            cohortids=cohortids,
        )
        return self._trs(CohortMembers, data)

    def get_cohorts(self, cohortids: List[int]) -> List[Cohort]:
        data = self.moodle.get('core_cohort_get_cohorts', cohortids=cohortids)
        return self._trs(Cohort, data)

    def search_cohorts(
        self,
        query: str,
        contextid: int = 0,
        contextlevel: str = "",
        instanceid: int = 0,
        includes: str = "parents",
        limitfrom: int = 0,
        limitnum: int = 25,
    ) -> Cohorts:
        """Search for cohorts.

        Args:
            query (str): Query string
            contextid (int, optional): Context ID. Either use this value, or level and instanceid.. Defaults to 0.
            contextlevel (str, optional): Context level. To be used with instanceid.. Defaults to "".
            instanceid (int, optional): Context instance ID. To be used with level. Defaults to 0.
            includes (str, optional): What other contexts to fetch the frameworks from. (all, parents, self). Defaults to "parents".
            limitfrom (int, optional): limitfrom we are fetching the records from. Defaults to 0.
            limitnum (int, optional): Number of records to fetch. Defaults to 25.

        Returns:
            Cohorts: list of Cohort
        """
        context = {
            "contextid": contextid,
            "contextlevel": contextlevel,
            "instanceid": instanceid,
        }
        data = self.moodle.get(
            'core_cohort_search_cohorts',
            query=query,
            context=context,
            includes=includes,
            limitfrom=limitfrom,
            limitnum=limitnum,
        )
        return self._tr(Cohorts, **data)

    def update_cohorts(self, cohorts: List[UpdateCohort]) -> None:
        """Updates existing cohorts.

        Args:
            cohorts (List[UpdateCohort]): list of UpdateCohort
        """
        self.moodle.get('core_cohort_update_cohorts', cohorts=cohorts)



from typing import List, Optional

from moodle import ResponsesFactory
from moodle.attr import dataclass, field


@dataclass
class Timestamp:
    """Timestamp

    Args:
        key (str): Timestamp key
        timestamp (int): Unix timestamp
    """
    key: str
    timestamp: int


@dataclass
class Timestamps(ResponsesFactory[Timestamp]):
    """list of Timestamp
    """
    timestamps: List[Timestamp] = field(factory=list)

    @property
    def items(self) -> List[Timestamp]:
        return self.timestamps

    @dataclass
    class Create:
        """Arg for core_calendar_get_timestamps

        Args:
            key (Optional[str]): key
            year (int): year
            month (int): month
            day (int): day
            hour (Optional[int]): hour
            minute (Optional[int]): minute
        """
        key: Optional[str]
        year: int
        month: int
        day: int
        hour: Optional[int]
        minute: Optional[int]



from moodle.base.general import GeneralTypeValue


class UserType(GeneralTypeValue):
    """Cohort Usertype

    Args:
        type (str): The name of the field: id
                        (numeric value of id) or username (alphanumeric value of username)
        value (str): The value of the cohort
    """
    pass
